/** fMSX: portable MSX emulator ******************************/
/**                                                         **/
/**                         Common.h                        **/
/**                                                         **/
/** This file contains standard screen refresh drivers      **/
/** common for X11, VGA, and other "chunky" bitmapped video **/
/** implementations. It also includes basic disk drivers    **/
/** and dummy sound drivers for fMSX.                       **/
/**                                                         **/
/** Copyright (C) Marat Fayzullin 1994-1998                 **/
/**               John Stiles     1996                      **/
/**     You are not allowed to distribute this software     **/
/**     commercially. Please, notify me, if you make any    **/
/**     changes to this file.                               **/
/*************************************************************/

static void Sprites(byte Y);
static void ColorSprites(byte Y,byte *ZBuf);
static byte *RefreshBorder(byte Y,byte C);
static word *RefreshBorder16bpp(byte Y,word C);
static void memset16bpp(word *P, word color, int cnt);

/** memset16bpp() ********************************************/
/* function called to set 16 bits values to memory           */
/*************************************************************/
void memset16bpp(word *P, word color, int cnt)
{
  register int r_cnt=cnt;
  register word *r_P=P;
  register word r_color=color;

  while (r_cnt--)
    *r_P++ = r_color;
}

/** DiskPresent() *********************************************/
/** Return 1 if disk drive with a given ID is present.       **/
/**************************************************************/
byte DiskPresent(byte ID)
{ return((ID<2)&&(Drives[ID]>=0)); }

/** DiskRead() ************************************************/
/** Read requested sector from the drive into a buffer.      **/
/**************************************************************/
byte DiskRead(byte ID,byte *Buf,int N)
{
  if((ID<2)&&(Drives[ID]>=0))
    if(lseek(Drives[ID],N*512L,0)==N*512L)
      return(read(Drives[ID],Buf,512)==512);
  return(0);
}

/** DiskWrite() ***********************************************/
/** Write contents of the buffer into a given sector of the  **/
/** disk.                                                    **/
/**************************************************************/
byte DiskWrite(byte ID,byte *Buf,int N)
{
  if((ID<2)&&(Drives[ID]>=0))
    if(lseek(Drives[ID],N*512L,0)==N*512L)
      return(write(Drives[ID],Buf,512)==512);
  return(0);
}

/** RefreshScreen() ******************************************/
/** Refresh screen. This function is called in the end of   **/
/** refresh cycle to show the entire screen.                **/
/*************************************************************/
void RefreshScreen(void) { PutImage(); }

/** RefreshBorder() ******************************************/
/** This function is called from RefreshLine#() to refresh  **/
/** the screen border. It returns a pointer to the start of **/
/** scanline Y in XBuf.                                     **/
/*************************************************************/
byte *RefreshBorder(register byte Y,register byte C)
{
  register byte H,*P;

  /* Set up the transparent color */
  XPal[0]=(!BGColor||SolidColor0)? XPal0:XPal[BGColor];

  /* Start of the line */
  P=XBuf+WIDTH*(Y+8);

  /* Width of a border */
  H=(WIDTH-256)/2;

  /* Paint left/right borders */
  memset(P,C,H);memset(P+H+256,C,H);

  /* Paint top of the border */
  if(!Y) memset(XBuf,C,WIDTH*8);

  /* Paint bottom of the border */
  H=ScanLines212? 212:192;
  if(Y==H-1) memset(P+WIDTH,C,WIDTH*(HEIGHT-H-8));

  /* Return pointer to the scanline in XBuf */
  return(P+(WIDTH-256)/2);
}

/** RefreshBorder16bpp() *************************************/
/** This function is called from RefreshLine#() to refresh  **/
/** the screen border. It returns a pointer to the start of **/
/** scanline Y in XBuf.                                     **/
/*************************************************************/
word *RefreshBorder16bpp(register byte Y,register word C)
{
  register byte H;
  register word *P;

  /* Set up the transparent color */
  XPal16bpp[0]=(!BGColor||SolidColor0)? XPal16bpp0:XPal16bpp[BGColor];

  /* Start of the line */
  P=((word *)XBuf)+WIDTH*(Y+8);

  /* Width of a border */
  H=(WIDTH-256)/2;

  /* Paint left/right borders */
  memset16bpp(P,C,H);memset16bpp(P+H+256,C,H);

  /* Paint top of the border */
  if(!Y) memset16bpp((word *)XBuf,C,WIDTH*8);

  /* Paint bottom of the border */
  H=ScanLines212? 212:192;
  if(Y==H-1) memset16bpp(P+WIDTH,C,WIDTH*(HEIGHT-H-8));

  /* Return pointer to the scanline in XBuf */
  return(P+(WIDTH-256)/2);
}

/** RefreshLineF() *******************************************/
/** Dummy refresh function called for non-existing screens. **/
/*************************************************************/
void RefreshLineF(register byte Y)
{
  if(Verbose>1)
    printf
    (
      "ScrMODE %d: ChrTab=%X ChrGen=%X ColTab=%X SprTab=%X SprGen=%X\n",
      ScrMode,ChrTab-VRAM,ChrGen-VRAM,ColTab-VRAM,SprTab-VRAM,SprGen-VRAM
    );

  memset(RefreshBorder(Y,XPal[BGColor]),XPal[BGColor],256);
}

/** RefreshLine0() *******************************************/
/** Refresh line Y (0..191/211) of SCREEN0.                 **/
/*************************************************************/
void RefreshLine0(register byte Y)
{
  register byte X;
  register byte *T,*G;

  if (imgdepth != 16) {
    register byte FC,BC;
    register byte *P;

    BC=XPal[BGColor];
    P=RefreshBorder(Y,BC);
  
    if(!ScreenON) memset(P,BC,256);
    else
    {
      P[0]=P[1]=P[2]=P[3]=P[4]=P[5]=P[6]=P[7]=BC;
      G=(UseFont&&FontBuf? FontBuf:ChrGen)+(Y&0x07);
      T=ChrTab+40*(Y>>3);
      FC=XPal[FGColor];
      P+=8;
  
      for(X=0;X<40;X++,T++,P+=6)
      {
        Y=G[(int)*T<<3];
        P[0]=Y&0x80? FC:BC;P[1]=Y&0x40? FC:BC;
        P[2]=Y&0x20? FC:BC;P[3]=Y&0x10? FC:BC;
        P[4]=Y&0x08? FC:BC;P[5]=Y&0x04? FC:BC;
      }
  
      P[0]=P[1]=P[2]=P[3]=P[4]=P[5]=P[6]=P[7]=BC;
    }
  }
  else {
    register word FC,BC;
    register word *P;

    BC=XPal16bpp[BGColor];
    P=RefreshBorder16bpp(Y,BC);
  
    if(!ScreenON) memset16bpp(P,BC,256);
    else
    {
      P[0]=P[1]=P[2]=P[3]=P[4]=P[5]=P[6]=P[7]=BC;
      G=(UseFont&&FontBuf? FontBuf:ChrGen)+(Y&0x07);
      T=ChrTab+40*(Y>>3);
      FC=XPal16bpp[FGColor];
      P+=8;
  
      for(X=0;X<40;X++,T++,P+=6)
      {
        Y=G[(int)*T<<3];
        P[0]=Y&0x80? FC:BC;P[1]=Y&0x40? FC:BC;
        P[2]=Y&0x20? FC:BC;P[3]=Y&0x10? FC:BC;
        P[4]=Y&0x08? FC:BC;P[5]=Y&0x04? FC:BC;
      }
  
      P[0]=P[1]=P[2]=P[3]=P[4]=P[5]=P[6]=P[7]=BC;
    }
  }
}

/** RefreshLine1() *******************************************/
/** Refresh line Y (0..191/211) of SCREEN1, including       **/
/** sprites in this line.                                   **/
/*************************************************************/
void RefreshLine1(register byte Y)
{
  register byte K,X;
  register byte *T,*G;

  if (imgdepth != 16) {
    register byte FC,BC;
    register byte *P;

    P=RefreshBorder(Y,XPal[BGColor]);
  
    if(!ScreenON) memset(P,XPal[BGColor],256);
    else
    {
      G=(UseFont&&FontBuf? FontBuf:ChrGen)+(Y&0x07);
      T=ChrTab+((int)(Y&0xF8)<<2);
  
      for(X=0;X<32;X++,T++,P+=8)
      {
        K=ColTab[*T>>3];
        FC=XPal[K>>4];
        BC=XPal[K&0x0F];
        K=G[(int)*T<<3];
        P[0]=K&0x80? FC:BC;P[1]=K&0x40? FC:BC;
        P[2]=K&0x20? FC:BC;P[3]=K&0x10? FC:BC;
        P[4]=K&0x08? FC:BC;P[5]=K&0x04? FC:BC;
        P[6]=K&0x02? FC:BC;P[7]=K&0x01? FC:BC;
      }
  
      if(!SpritesOFF) Sprites(Y);
    }
  }
  else {
    register word FC,BC;
    register word *P;

    P=RefreshBorder16bpp(Y,XPal16bpp[BGColor]);
  
    if(!ScreenON) memset16bpp(P,XPal16bpp[BGColor],256);
    else
    {
      G=(UseFont&&FontBuf? FontBuf:ChrGen)+(Y&0x07);
      T=ChrTab+((int)(Y&0xF8)<<2);
  
      for(X=0;X<32;X++,T++,P+=8)
      {
        K=ColTab[*T>>3];
        FC=XPal16bpp[K>>4];
        BC=XPal16bpp[K&0x0F];
        K=G[(int)*T<<3];
        P[0]=K&0x80? FC:BC;P[1]=K&0x40? FC:BC;
        P[2]=K&0x20? FC:BC;P[3]=K&0x10? FC:BC;
        P[4]=K&0x08? FC:BC;P[5]=K&0x04? FC:BC;
        P[6]=K&0x02? FC:BC;P[7]=K&0x01? FC:BC;
      }
  
      if(!SpritesOFF) Sprites(Y);
    }
  }
}

/** RefreshLine2() *******************************************/
/** Refresh line Y (0..191/211) of SCREEN2, including       **/
/** sprites in this line.                                   **/
/*************************************************************/
void RefreshLine2(register byte Y)
{
  register byte K,X;
  register byte *T,*PGT,*CLT;
  register int J;

  if (imgdepth!=16) {
    register byte *P;
    register byte FC,BC;

    P=RefreshBorder(Y,XPal[BGColor]);
  
    if(!ScreenON) memset(P,XPal[BGColor],256);
    else
    {
      T=ChrTab+((int)(Y&0xF8)<<2);
      J=((int)(Y&0xC0)<<5)+(Y&0x07);
      PGT=ChrGen+J;
      CLT=ColTab+J;
  
      for(X=0;X<32;X++,T++,P+=8)
      {
        J=(int)*T<<3;
        K=CLT[J];
        FC=XPal[K>>4];
        BC=XPal[K&0x0F];
        K=PGT[J];
        P[0]=K&0x80? FC:BC;P[1]=K&0x40? FC:BC;
        P[2]=K&0x20? FC:BC;P[3]=K&0x10? FC:BC;
        P[4]=K&0x08? FC:BC;P[5]=K&0x04? FC:BC;
        P[6]=K&0x02? FC:BC;P[7]=K&0x01? FC:BC;
      }
  
      if(!SpritesOFF) Sprites(Y);
    }
  }
  else {
    register word *P;
    register word FC,BC;

    P=RefreshBorder16bpp(Y,XPal16bpp[BGColor]);
  
    if(!ScreenON) memset16bpp(P,XPal16bpp[BGColor],256);
    else
    {
      T=ChrTab+((int)(Y&0xF8)<<2);
      J=((int)(Y&0xC0)<<5)+(Y&0x07);
      PGT=ChrGen+J;
      CLT=ColTab+J;
  
      for(X=0;X<32;X++,T++,P+=8)
      {
        J=(int)*T<<3;
        K=CLT[J];
        FC=XPal16bpp[K>>4];
        BC=XPal16bpp[K&0x0F];
        K=PGT[J];
        P[0]=K&0x80? FC:BC;P[1]=K&0x40? FC:BC;
        P[2]=K&0x20? FC:BC;P[3]=K&0x10? FC:BC;
        P[4]=K&0x08? FC:BC;P[5]=K&0x04? FC:BC;
        P[6]=K&0x02? FC:BC;P[7]=K&0x01? FC:BC;
      }
  
      if(!SpritesOFF) Sprites(Y);
    }
  }
}

/** RefreshLine3() *******************************************/
/** Refresh line Y (0..191/211) of SCREEN3, including       **/
/** sprites in this line.                                   **/
/*************************************************************/
void RefreshLine3(register byte Y)
{
  register byte X,K;
  register byte *T,*G;

  if (imgdepth != 16) {
    register byte *P;

    P=RefreshBorder(Y,XPal[BGColor]);
  
    if(!ScreenON) memset(P,XPal[BGColor],256);
    else
    {
      T=ChrTab+((int)(Y&0xF8)<<2);
      G=ChrGen+((Y&0x1C)>>2);
  
      for(X=0;X<32;X++,T++,P+=8)
      {
        K=G[(int)*T<<3];
        P[0]=P[1]=P[2]=P[3]=XPal[K>>4];
        P[4]=P[5]=P[6]=P[7]=XPal[K&0x0F];
      }
    }
  }
  else {
    register word *P;

    P=RefreshBorder16bpp(Y,XPal16bpp[BGColor]);
  
    if(!ScreenON) memset16bpp(P,XPal16bpp[BGColor],256);
    else
    {
      T=ChrTab+((int)(Y&0xF8)<<2);
      G=ChrGen+((Y&0x1C)>>2);
  
      for(X=0;X<32;X++,T++,P+=8)
      {
        K=G[(int)*T<<3];
        P[0]=P[1]=P[2]=P[3]=XPal16bpp[K>>4];
        P[4]=P[5]=P[6]=P[7]=XPal16bpp[K&0x0F];
      }
    }
  }
}

/** RefreshLine4() *******************************************/
/** Refresh line Y (0..191/211) of SCREEN4, including       **/
/** sprites in this line.                                   **/
/*************************************************************/
void RefreshLine4(register byte Y)
{
  register byte I,X,C;
  register byte *T,*R;
  register int L;
  byte ZBuf[304];

  if (imgdepth != 16) {
    register byte *P;
    register byte BC,FC;

    P=RefreshBorder(Y,XPal[BGColor]);
  
    if(!ScreenON) memset(P,XPal[BGColor],256);
    else
    {
      ColorSprites(Y,ZBuf);
      R=ZBuf+32;
      Y+=VDP[23];
      T=ChrTab+((int)(Y&0xF8)<<2);
  
      for(X=0;X<32;X++,R+=8,P+=8,T++)
      {
        L=((int)(Y&0xC0)<<5)+((int)*T<<3)+(Y&0x07);
        I=ColTab[L];
        FC=XPal[I>>4];
        BC=XPal[I&0x0F];
        I=ChrGen[L];
  
        C=R[0];P[0]=C? XPal[C]:(I&0x80)? FC:BC;
        C=R[1];P[1]=C? XPal[C]:(I&0x40)? FC:BC;
        C=R[2];P[2]=C? XPal[C]:(I&0x20)? FC:BC;
        C=R[3];P[3]=C? XPal[C]:(I&0x10)? FC:BC;
        C=R[4];P[4]=C? XPal[C]:(I&0x08)? FC:BC;
        C=R[5];P[5]=C? XPal[C]:(I&0x04)? FC:BC;
        C=R[6];P[6]=C? XPal[C]:(I&0x02)? FC:BC;
        C=R[7];P[7]=C? XPal[C]:(I&0x01)? FC:BC;
      }
    }
  }
  else {
    register word *P;
    register word BC,FC;

    P=RefreshBorder16bpp(Y,XPal16bpp[BGColor]);
  
    if(!ScreenON) memset16bpp(P,XPal16bpp[BGColor],256);
    else
    {
      ColorSprites(Y,ZBuf);
      R=ZBuf+32;
      Y+=VDP[23];
      T=ChrTab+((int)(Y&0xF8)<<2);
  
      for(X=0;X<32;X++,R+=8,P+=8,T++)
      {
        L=((int)(Y&0xC0)<<5)+((int)*T<<3)+(Y&0x07);
        I=ColTab[L];
        FC=XPal16bpp[I>>4];
        BC=XPal16bpp[I&0x0F];
        I=ChrGen[L];
  
        C=R[0];P[0]=C? XPal16bpp[C]:(I&0x80)? FC:BC;
        C=R[1];P[1]=C? XPal16bpp[C]:(I&0x40)? FC:BC;
        C=R[2];P[2]=C? XPal16bpp[C]:(I&0x20)? FC:BC;
        C=R[3];P[3]=C? XPal16bpp[C]:(I&0x10)? FC:BC;
        C=R[4];P[4]=C? XPal16bpp[C]:(I&0x08)? FC:BC;
        C=R[5];P[5]=C? XPal16bpp[C]:(I&0x04)? FC:BC;
        C=R[6];P[6]=C? XPal16bpp[C]:(I&0x02)? FC:BC;
        C=R[7];P[7]=C? XPal16bpp[C]:(I&0x01)? FC:BC;
      }
    }
  }
}

/** RefreshLine5() *******************************************/
/** Refresh line Y (0..191/211) of SCREEN5, including       **/
/** sprites in this line.                                   **/
/*************************************************************/
void RefreshLine5(register byte Y)
{
  register byte I,X;
  register byte *T,*R;
  byte ZBuf[304];

  if (imgdepth != 16) {
    register byte *P;

    P=RefreshBorder(Y,XPal[BGColor]);
  
    if(!ScreenON) memset(P,XPal[BGColor],256);
    else
    {
      ColorSprites(Y,ZBuf);
      R=ZBuf+32;
      Y+=VDP[23];
      T=ChrTab+((int)Y<<7);
  
      for(X=0;X<16;X++,R+=16,P+=16,T+=8)
      {
        I=R[0];P[0]=XPal[I? I:T[0]>>4];
        I=R[1];P[1]=XPal[I? I:T[0]&0x0F];
        I=R[2];P[2]=XPal[I? I:T[1]>>4];
        I=R[3];P[3]=XPal[I? I:T[1]&0x0F];
        I=R[4];P[4]=XPal[I? I:T[2]>>4];
        I=R[5];P[5]=XPal[I? I:T[2]&0x0F];
        I=R[6];P[6]=XPal[I? I:T[3]>>4];
        I=R[7];P[7]=XPal[I? I:T[3]&0x0F];
        I=R[8];P[8]=XPal[I? I:T[4]>>4];
        I=R[9];P[9]=XPal[I? I:T[4]&0x0F];
        I=R[10];P[10]=XPal[I? I:T[5]>>4];
        I=R[11];P[11]=XPal[I? I:T[5]&0x0F];
        I=R[12];P[12]=XPal[I? I:T[6]>>4];
        I=R[13];P[13]=XPal[I? I:T[6]&0x0F];
        I=R[14];P[14]=XPal[I? I:T[7]>>4];
        I=R[15];P[15]=XPal[I? I:T[7]&0x0F];
      }
    }
  }
  else {
    register word *P;

    P=RefreshBorder16bpp(Y,XPal16bpp[BGColor]);
  
    if(!ScreenON) memset16bpp(P,XPal16bpp[BGColor],256);
    else
    {
      ColorSprites(Y,ZBuf);
      R=ZBuf+32;
      Y+=VDP[23];
      T=ChrTab+((int)Y<<7);
  
      for(X=0;X<16;X++,R+=16,P+=16,T+=8)
      {
        I=R[0];P[0]=XPal16bpp[I? I:T[0]>>4];
        I=R[1];P[1]=XPal16bpp[I? I:T[0]&0x0F];
        I=R[2];P[2]=XPal16bpp[I? I:T[1]>>4];
        I=R[3];P[3]=XPal16bpp[I? I:T[1]&0x0F];
        I=R[4];P[4]=XPal16bpp[I? I:T[2]>>4];
        I=R[5];P[5]=XPal16bpp[I? I:T[2]&0x0F];
        I=R[6];P[6]=XPal16bpp[I? I:T[3]>>4];
        I=R[7];P[7]=XPal16bpp[I? I:T[3]&0x0F];
        I=R[8];P[8]=XPal16bpp[I? I:T[4]>>4];
        I=R[9];P[9]=XPal16bpp[I? I:T[4]&0x0F];
        I=R[10];P[10]=XPal16bpp[I? I:T[5]>>4];
        I=R[11];P[11]=XPal16bpp[I? I:T[5]&0x0F];
        I=R[12];P[12]=XPal16bpp[I? I:T[6]>>4];
        I=R[13];P[13]=XPal16bpp[I? I:T[6]&0x0F];
        I=R[14];P[14]=XPal16bpp[I? I:T[7]>>4];
        I=R[15];P[15]=XPal16bpp[I? I:T[7]&0x0F];
      }
    }
  }
}

/** Sprites() ************************************************/
/** This function is called from RefreshLine#() to refresh  **/
/** sprites in SCREENs 1-3.                                 **/
/*************************************************************/
void Sprites(register byte Y)
{
  register byte C,H;
  register byte *P,*PT,*AT;
  register int L,K;
  register unsigned int M;

  /* Assign initial values before counting */
  H=Sprites16x16? 16:8;
  C=0;M=0;L=0;
  AT=SprTab-4;

  /* Count displayed sprites */
  do
  {
    M<<=1;AT+=4;L++;    /* Iterating through SprTab      */
    K=AT[0];            /* K = sprite Y coordinate       */
    if(K==208) break;   /* Iteration terminates if Y=208 */
    if(K>256-H) K-=256; /* Y coordinate may be negative  */

    /* Mark all valid sprites with 1s, break at 4 sprites */
    if((Y>K)&&(Y<=K+H)) { M|=1;C++;if(C==4) break; }
  }
  while(L<32);

  /* Draw all marked sprites */
  for(;M;M>>=1,AT-=4)
    if(M&1)
    {
      C=AT[3];                  /* C = sprite attributes */
      L=C&0x80? AT[1]-32:AT[1]; /* Sprite may be shifted left by 32 */
      C&=0x0F;                  /* C = sprite color */

      if((L<256)&&(L>-H)&&C)
      {
        K=AT[0];                /* K = sprite Y coordinate */
        if(K>256-H) K-=256;     /* Y coordinate may be negative */

        P=XBuf+(WIDTH-256)/2+WIDTH*(Y+8)+L;
        PT=SprGen+((int)(H>8? AT[2]&0xFC:AT[2])<<3)+Y-K-1;
        C=XPal[C];

        /* Mask 1: clip left sprite boundary */
        K=L>=0? 0x0FFFF:(0x10000>>-L)-1;
        /* Mask 2: clip right sprite boundary */
        if(L>256-H) K^=((0x00200>>(H-8))<<(L-257+H))-1;
        /* Get and clip the sprite data */
        K&=((int)PT[0]<<8)|(H>8? PT[16]:0x00);

        /* Draw left 8 pixels of the sprite */
        if(K&0xFF00)
        {
          if(K&0x8000) P[0]=C;if(K&0x4000) P[1]=C;
          if(K&0x2000) P[2]=C;if(K&0x1000) P[3]=C;
          if(K&0x0800) P[4]=C;if(K&0x0400) P[5]=C;
          if(K&0x0200) P[6]=C;if(K&0x0100) P[7]=C;
        }

        /* Draw right 8 pixels of the sprite */
        if(K&0x00FF)
        {
          if(K&0x0080) P[8]=C; if(K&0x0040) P[9]=C;
          if(K&0x0020) P[10]=C;if(K&0x0010) P[11]=C;
          if(K&0x0008) P[12]=C;if(K&0x0004) P[13]=C;
          if(K&0x0002) P[14]=C;if(K&0x0001) P[15]=C;
        }
      }
    }
}

/** ColorSprites() *******************************************/
/** This function is called from RefreshLine#() to refresh  **/
/** color sprites in SCREENs 4-8. The result is returned in **/
/** ZBuf, whose size must be 304 bytes (32+256+16).         **/
/*************************************************************/
void ColorSprites(register byte Y,byte *ZBuf)
{
  register byte C,H,J,OrThem;
  register byte *P,*PT,*AT;
  register int L,K;
  register unsigned int M;

  /* Clear ZBuffer and exit if sprites are off */
  memset(ZBuf+32,0,256);
  if(SpritesOFF) return;

  /* Assign initial values before counting */
  H=Sprites16x16? 16:8;
  C=0;M=0;L=0;
  AT=SprTab-4;
  OrThem=0x00;

  /* Count displayed sprites */
  do
  {
    M<<=1;AT+=4;L++;          /* Iterating through SprTab      */
    K=AT[0];                  /* Read Y from SprTab            */
    if(K==216) break;         /* Iteration terminates if Y=216 */
    K=(byte)(K-VDP[23]);      /* Sprite's actual Y coordinate  */
    if(K>256-H) K-=256;       /* Y coordinate may be negative  */

    /* Mark all valid sprites with 1s, break at 8 sprites */
    if((Y>K)&&(Y<=K+H)) { M|=1;C++;if(C==8) break; }
  }
  while(L<32);

  /* Draw all marked sprites */
  for(;M;M>>=1,AT-=4)
    if(M&1)
    {
      K=(byte)(AT[0]-VDP[23]); /* K = sprite Y coordinate */
      if(K>256-H) K-=256;      /* Y coordinate may be negative */

      J=Y-K-1;
      PT=SprGen+((int)(H>8? AT[2]&0xFC:AT[2])<<3)+J;
      C=SprTab[-0x0200+((AT-SprTab)<<2)+J];
      P=ZBuf+AT[1]+(C&0x80? 0:32);
      OrThem|=C&0x40;
      C&=0x0F;
      J=PT[0];

      if(OrThem&0x20)
      {
        if(J&0x80) P[0]|=C;if(J&0x40) P[1]|=C;
        if(J&0x20) P[2]|=C;if(J&0x10) P[3]|=C;
        if(J&0x08) P[4]|=C;if(J&0x04) P[5]|=C;
        if(J&0x02) P[6]|=C;if(J&0x01) P[7]|=C;
        if(H>8)
        {
          J=PT[16];
          if(J&0x80) P[8]|=C; if(J&0x40) P[9]|=C;
          if(J&0x20) P[10]|=C;if(J&0x10) P[11]|=C;
          if(J&0x08) P[12]|=C;if(J&0x04) P[13]|=C;
          if(J&0x02) P[14]|=C;if(J&0x01) P[15]|=C;
        }
      }
      else
      {
        if(J&0x80) P[0]=C;if(J&0x40) P[1]=C;
        if(J&0x20) P[2]=C;if(J&0x10) P[3]=C;
        if(J&0x08) P[4]=C;if(J&0x04) P[5]=C;
        if(J&0x02) P[6]=C;if(J&0x01) P[7]=C;
        if(H>8)
        {
          J=PT[16];
          if(J&0x80) P[8]=C; if(J&0x40) P[9]=C;
          if(J&0x20) P[10]=C;if(J&0x10) P[11]=C;
          if(J&0x08) P[12]=C;if(J&0x04) P[13]=C;
          if(J&0x02) P[14]=C;if(J&0x01) P[15]=C;
        }
      }

      /* Update overlapping flag */
      OrThem>>=1;
    }
}
