/** fMSX: portable MSX emulator ******************************/
/**                                                         **/
/**                           Unix.c                        **/
/**                                                         **/
/** This file contains Unix/X-dependent subroutines and     **/
/** drivers. It includes common drivers from Common.h.      **/
/**                                                         **/
/** Copyright (C) Marat Fayzullin 1994-1998                 **/
/**               Elan Feingold   1995                      **/
/**               Ville Hallik    1996                      **/
/**     You are not allowed to distribute this software     **/
/**     commercially. Please, notify me, if you make any    **/
/**     changes to this file.                               **/
/*************************************************************/
#ifdef UNIX

/** Private #includes ****************************************/
#include "MSX.h"
#include "LibUnix.h"

/** Standard Unix/X #includes ********************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/keysym.h>

/** Public parameters ****************************************/
char *Title="fMSX Unix/X 2.0";  /* Window/command line title */
int SaveCPU  = 1;               /* 1 = freeze when focus out */
int UseSHM   = 1;               /* 1 = use MITSHM            */
int IFreq=50;

/** Various variables ****************************************/
#define WIDTH  340
#define HEIGHT 240

static byte *XBuf;
static byte BPal[256],XPal[16],XPal0; 
static word BPal16bpp[256],XPal16bpp[16],XPal16bpp0;
static Image Img; 
static byte JoyState;
static int imgdepth;

/** Various X-related variables ******************************/
static Display *Dsp;
static Window Wnd;
static Colormap DefaultCMap;
static GC DefaultGC;
static unsigned long White,Black;

/** Sound-related definitions ********************************/
#ifdef SOUND
int InitSound(void);
void TrashSound(void);
void FlushSound(void);
void StopSound(void);
void ResumeSound(void);
extern int UseSound;
#endif

/** This function is called on signals ***********************/
static void OnBreak(int Arg) { ExitNow=1;signal(Arg,OnBreak); }

/** Screen Mode Handlers [number of screens + 1] *************/
void (*RefreshLine[MAXSCREEN+2])(byte Y) =
{
  RefreshLine0,   /* SCR 0:TEXT 40x24  */
  RefreshLine1,   /* SCR 1:TEXT 32x24  */
  RefreshLine2,   /* SCR 2:BLK 256x192 */
  RefreshLine3,   /* SCR 3:64x48x16    */
  RefreshLine4,   /* SCR 4:BLK 256x192 */
  RefreshLine5,   /* SCR 5:256x192x16  */
  RefreshLine6,   /* SCR 6:512x192x4   */
  RefreshLine7,   /* SCR 7:512x192x16  */
  RefreshLine8,   /* SCR 8:256x192x256 */
  RefreshLineTx80 /* SCR 0:TEXT 80x24  */
};

/** Keyboard bindings ****************************************/
byte Keys[256][2] =
{
/* 00 */
  {0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},
  {7,0x20},{7,0x08},{0,0x00},{0,0x00},{0,0x00},{7,0x80},{0,0x00},{0,0x00},
/* 10 */
  {0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},
  {0,0x00},{0,0x00},{0,0x00},{7,0x04},{0,0x00},{0,0x00},{0,0x00},{0,0x00},
/* 20 */
  {8,0x01},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{2,0x01},
  {0,0x00},{0,0x00},{0,0x00},{0,0x00},{2,0x04},{1,0x04},{2,0x08},{2,0x10},
/* 30 */
  {0,0x01},{0,0x02},{0,0x04},{0,0x08},{0,0x10},{0,0x20},{0,0x40},{0,0x80},
  {1,0x01},{1,0x02},{0,0x00},{1,0x80},{0,0x00},{1,0x08},{0,0x00},{0,0x00},
/* 40 */
  {0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},
  {0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},
/* 50 */
  {0,0x00},{8,0x10},{8,0x20},{8,0x80},{8,0x40},{0,0x00},{0,0x00},{0,0x00},
  {0,0x00},{0,0x00},{0,0x00},{1,0x20},{1,0x10},{1,0x40},{0,0x00},{0,0x00},
/* 60 */
  {2,0x02},{2,0x40},{2,0x80},{3,0x01},{3,0x02},{3,0x04},{3,0x08},{3,0x10},
  {3,0x20},{3,0x40},{3,0x80},{4,0x01},{4,0x02},{4,0x04},{4,0x08},{4,0x10},
/* 70 */
  {4,0x20},{4,0x40},{4,0x80},{5,0x01},{5,0x02},{5,0x04},{5,0x08},{5,0x10},
  {5,0x20},{5,0x40},{5,0x80},{0,0x00},{0,0x00},{0,0x00},{6,0x04},{0,0x00},
/* 80 */
  {0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},
  {0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},
/* 90 */
  {0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},
  {0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},
/* A0 */
  {0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},
  {0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},
/* B0 */
  {0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},
  {0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{6,0x20},{6,0x40},
/* C0 */
  {6,0x80},{7,0x01},{7,0x02},{8,0x02},{8,0x08},{8,0x04},{7,0x40},{7,0x10},
  {0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},
/* D0 */
  {0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},
  {8,0x02},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},
/* E0 */
  {0,0x00},{6,0x01},{6,0x08},{6,0x02},{0,0x00},{0,0x00},{0,0x00},{0,0x00},
  {0,0x00},{6,0x10},{6,0x04},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},
/* F0 */
  {0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},
  {0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{8,0x08}
};

/** InitMachine() ********************************************/
/** Allocate resources needed by Unix/X-dependent code.     **/
/*************************************************************/
int InitMachine(void)
{
  Screen *Scr;
  int J;

  /* Reset the variables */
  XBuf=NULL;

  /* Open display */
  if(Verbose) printf("Initializing Unix/X drivers:\n  Opening display...");
  if(!(Dsp=XOpenDisplay(NULL))) { if(Verbose) puts("FAILED");return(0); }

  /* Set internal variables */
  Scr=DefaultScreenOfDisplay(Dsp);
  White=WhitePixelOfScreen(Scr);
  Black=BlackPixelOfScreen(Scr);
  DefaultGC=DefaultGCOfScreen(Scr);
  DefaultCMap=DefaultColormapOfScreen(Scr);

  /* Reset the palette */
  XPal0=Black;
  XPal16bpp0=0;
  for(J=0;J<16;J++) {
    XPal[J]=Black;
    XPal16bpp[J]=0;
  }

  /* Keys do not autorepeat */
  XAutoRepeatOff(Dsp);

  /* No joystick buttons for now */
  JoyState=0x00;

  /* Initialize LibUnix toolkit */
  if(Verbose) printf("OK\n  Initializing LibUnix...");
  if(!InitLibUnix(Dsp)) { if(Verbose) puts("FAILED");return(0); }
  
  /* Create a window */
  if(Verbose) printf("OK\n  Opening window...");
  if(!(Wnd=X11Window(Title,WIDTH,HEIGHT)))
  { if(Verbose) puts("FAILED");return(0); }
  
  /* Create an image */
  if(Verbose) printf("OK\n  Allocating screen buffer...");
  if(!X11NewImage(&Img,WIDTH,HEIGHT,UseSHM? USE_SHM:0))
  { if(Verbose) puts("FAILED");return(0); }
  XBuf=Img.Data;
  imgdepth=Img.D;

  /* Done allocating screen buffer */ 
  puts("OK");

  /* Allocate colors for SCREEN8 */
  if (imgdepth != 16) {
    XColor Color;
    word R,G,B;
    byte J,I;

    for(R=0;R<4;R++)
      for(G=0;G<4;G++)
        for(B=0;B<4;B++)
        {
          Color.flags=DoRed|DoGreen|DoBlue;
          Color.red=R<<14;
          Color.green=G<<14;
          Color.blue=B<<14;
          J=XAllocColor(Dsp,DefaultCMap,&Color)? Color.pixel:Black;
          I=B|(R<<3)|(G<<6);
          BPal[I]=BPal[I|0x04]=BPal[I|0x20]=BPal[I|0x24]=J;
        }  
  }
  else {
    XColor Color;
    word R,G,B;
    word J;

    for(R=0;R<8;R++)
      for(G=0;G<8;G++)
        for(B=0;B<4;B++) {
          Color.flags=DoRed|DoGreen|DoBlue;
          Color.red=(R<<13)|(R<<10)|(R<<7)|(R<<4)|(R<<1)|(R>>2);
          Color.green=(G<<13)|(G<<10)|(G<<7)|(G<<4)|(G<<1)|(G>>2);
          Color.blue=(B<<14)|(B<<12)|(B<<10)|(B<<8);
          Color.blue|=Color.blue>>8;
          J=XAllocColor(Dsp,DefaultCMap,&Color)? Color.pixel:0;
/*          printf("bpal (%d, %d, %d) -> %d\n", R, G, B, J); */
          BPal16bpp[(R<<2)|(G<<5)|B]=J;
        }
  }

#ifdef SOUND
  /* Initialize sound */
  if(UseSound) InitSound();
#endif SOUND

  /* Catch all signals */
  signal(SIGHUP,OnBreak);signal(SIGINT,OnBreak);
  signal(SIGQUIT,OnBreak);signal(SIGTERM,OnBreak);

  return(1);
}

/** TrashMachine() *******************************************/
/** Deallocate all resources taken by InitMachine().        **/
/*************************************************************/
void TrashMachine(void)
{
  unsigned long C;
  int J;

  if(Verbose) printf("Shutting down...\n");

  if(Dsp)
  {
    /* Free SCREEN8 palette */
    if (imgdepth != 16) 
      for(J=0;J<64;J++) {
        C=BPal[(J&0x03)|((J&0x0C)<<1)|((J&0x30)<<2)];
        XFreeColors(Dsp,DefaultCMap,&C,1,0);
      }

    X11FreeImage(&Img);
    /*TrashLibUnix();*/
    XAutoRepeatOn(Dsp);
    XCloseDisplay(Dsp);
  }

#ifdef SOUND
  StopSound();TrashSound();
#endif SOUND
}

/** PutImage() ***********************************************/
/** Put an image on the screen.                             **/
/*************************************************************/
void PutImage(void)
{ X11PutImage(Wnd,&Img,0,0,0,0,WIDTH,HEIGHT); }

/** Keyboard() ***********************************************/
/** Check for keyboard events, parse them, and modify MSX   **/
/** keyboard matrix.                                        **/
/*************************************************************/
void Keyboard(void)
{
  XEvent E;
  int J;

#ifdef SOUND
  FlushSound();  /* Flush sound stream on each interrupt */
#endif

  /* Check for keypresses/keyreleases */
  if(XCheckWindowEvent(Dsp,Wnd,KeyPressMask|KeyReleaseMask,&E))
  {
    J=XLookupKeysym((XKeyEvent *)&E,0);
    if(E.type==KeyPress)
    {
      /* Special keys pressed... */
      switch(J)
      {
#ifdef DEBUG
        case XK_F11:   CPU.Trace=!CPU.Trace;break;
#endif
        case XK_F12:   ExitNow=1;break;
        case XK_F13:   UseFont=UseFont? 0:FontBuf? 1:0;break;
        case XK_Up:    JoyState|=0x01;break;
        case XK_Down:  JoyState|=0x02;break;
        case XK_Left:  JoyState|=0x04;break;
        case XK_Right: JoyState|=0x08;break;
        case XK_Alt_L: case XK_Alt_R:
          JoyState|=0x10;break;
        case XK_Control_L: case XK_Control_R:
          JoyState|=0x20;break;
      }

      /* Key pressed: reset bit in KeyMap */
      J&=0xFF;KeyMap[Keys[J][0]]&=~Keys[J][1];
    }
    else
    {
      /* Special keys released... */
      switch(J)
      {
        case XK_Up:    JoyState&=0xFE;break;
        case XK_Down:  JoyState&=0xFD;break;
        case XK_Left:  JoyState&=0xFB;break;
        case XK_Right: JoyState&=0xF7;break;
        case XK_Alt_L: case XK_Alt_R:
          JoyState&=0xEF;break;
        case XK_Control_L: case XK_Control_R:
          JoyState&=0xDF;break;
      }

      /* Key released: set bit in KeyMap */
      J&=0xFF;KeyMap[Keys[J][0]]|=Keys[J][1];
    }
  }

  /* If saving CPU and focus is out, sleep */
  for(E.type=0;XCheckWindowEvent(Dsp,Wnd,FocusChangeMask,&E););
  if(SaveCPU&&(E.type==FocusOut))
  {
    XAutoRepeatOn(Dsp);
#ifdef SOUND
    StopSound();
#endif
  
    do
      while(!XCheckWindowEvent(Dsp,Wnd,FocusChangeMask,&E)&&!ExitNow)
        sleep(1);
    while((E.type!=FocusIn)&&!ExitNow);

    XAutoRepeatOff(Dsp);
#ifdef SOUND
    ResumeSound();
#endif
  }
}

/** Joystick() ***********************************************/
/** Query position of a joystick connected to port N.       **/
/** Returns 0.0.F2.F1.R.L.D.U.                              **/
/************************************ TO BE WRITTEN BY USER **/
byte Joystick(register byte N) { return(JoyState); }

/** Mouse() **************************************************/
/** Query coordinates of a mouse connected to port N.       **/
/** Returns F2.F1.Y.Y.Y.Y.Y.Y.Y.Y.X.X.X.X.X.X.X.X.          **/
/************************************ TO BE WRITTEN BY USER **/
int Mouse(register byte N)
{
  Window Dummy;
  int X,Y,J,Buttons;

  /* Get mouse coordinates */
  if(XQueryPointer(Dsp,Wnd,&Dummy,&Dummy,&J,&J,&X,&Y,&Buttons))
  {
    X-=(WIDTH-256)/2;  X=X<0? 0:X>255? 255:X;
    Y-=(HEIGHT-192)/2; Y=Y<0? 0:Y>211? 211:Y;
    Buttons=(Buttons&Button1Mask? 0x10000:0x00000)
           |(Buttons&Button2Mask? 0x20000:0x00000)
           |(Buttons&Button3Mask? 0x20000:0x00000);
  }
  else X=Y=Buttons=0;

  return((Y<<8)|X|Buttons);
}

/** SetColor() ***********************************************/
/** Set color N (0..15) to R,G,B.                           **/
/*************************************************************/
void SetColor(register byte N,register byte R,register byte G,register byte B)
{
  if (imgdepth != 16) {
    XPal[0]=XPal0;

    /* Allocate a new color entry */
    XPal[N]=X11GetColor(R,G,B);

    XPal0=XPal[0];
    XPal[0]=(!BGColor||SolidColor0)? XPal0:XPal[BGColor];
  }
  else {
    XColor Color;
    word rw, gw, bw;
    word J;

    XPal16bpp[0]=XPal16bpp0;

    Color.flags=DoRed|DoGreen|DoBlue;
    rw=R; gw=G; bw=B;
    Color.red=(rw<<8)|(rw<<5)|(rw<<2)|(rw>>1)|(rw>>4)|(rw>>7);
    Color.green=(gw<<8)|(gw<<5)|(gw<<2)|(gw>>1)|(gw>>4)|(gw>>7);
    Color.blue=(bw<<8)|(bw<<5)|(bw<<2)|(bw>>1)|(bw>>4)|(bw>>7);
    J=XAllocColor(Dsp,DefaultCMap,&Color)? Color.pixel:0;
    /* printf("xpal (%d, %d, %d) -> %d\n", R, G, B, J); */
    XPal16bpp[N]=J;

    XPal16bpp0=XPal16bpp[0];
    XPal16bpp[0]=(!BGColor||SolidColor0)? XPal16bpp0:XPal16bpp[BGColor];
  }
}

/** Dummy Functions ******************************************/
/** These dummy functions are called when writes to sound   **/
/** registers occur. Replace them with the actual functions **/
/** generating sound.                                       **/
/*************************************************************/
#ifndef SOUND
void PSGOut(register byte R,register byte V)  { }
void SCCOut(register byte R,register byte V)  { }
#endif
void PPIOut(register byte Toggle)             { }
void OPLLOut(register byte R,register byte V) { }

/** Part of the code common for Unix/X and MSDOS drivers *****/ 
#include "Common.h"

/** RefreshScr6() ********************************************/
/** Function to be called to update SCREEN 6.               **/
/*************************************************************/
void RefreshLine6(register byte Y)
{
  register byte X,C;
  register byte *T,*R;
  byte ZBuf[304];

  if (imgdepth != 16) {
    register byte *P;

    P=RefreshBorder(Y,XPal[BGColor&0x03]);
  
    if(!ScreenON) memset(P,XPal[BGColor],256);
    else
    {
      ColorSprites(Y,ZBuf);
      R=ZBuf+32;
      T=ChrTab+(((int)Y+VDP[23])<<7);
  
      for(X=0;X<32;X++)
      {
        C=R[0];P[0]=XPal[C? C:T[0]>>6];
        C=R[1];P[1]=XPal[C? C:(T[0]>>2)&0x03];
        C=R[2];P[2]=XPal[C? C:T[1]>>6];
        C=R[3];P[3]=XPal[C? C:(T[1]>>2)&0x03];
        C=R[4];P[4]=XPal[C? C:T[2]>>6];
        C=R[5];P[5]=XPal[C? C:(T[2]>>2)&0x03];
        C=R[6];P[6]=XPal[C? C:T[3]>>6];
        C=R[7];P[7]=XPal[C? C:(T[3]>>2)&0x03];
        R+=8;P+=8;T+=4;
      }
    }
  }
  else {
    register word *P;

    P=RefreshBorder16bpp(Y,XPal16bpp[BGColor&0x03]);
  
    if(!ScreenON) memset16bpp(P,XPal16bpp[BGColor],256);
    else
    {
      ColorSprites(Y,ZBuf);
      R=ZBuf+32;
      T=ChrTab+(((int)Y+VDP[23])<<7);
  
      for(X=0;X<32;X++)
      {
        C=R[0];P[0]=XPal16bpp[C? C:T[0]>>6];
        C=R[1];P[1]=XPal16bpp[C? C:(T[0]>>2)&0x03];
        C=R[2];P[2]=XPal16bpp[C? C:T[1]>>6];
        C=R[3];P[3]=XPal16bpp[C? C:(T[1]>>2)&0x03];
        C=R[4];P[4]=XPal16bpp[C? C:T[2]>>6];
        C=R[5];P[5]=XPal16bpp[C? C:(T[2]>>2)&0x03];
        C=R[6];P[6]=XPal16bpp[C? C:T[3]>>6];
        C=R[7];P[7]=XPal16bpp[C? C:(T[3]>>2)&0x03];
        R+=8;P+=8;T+=4;
      }
    }
  }

}
  
/** RefreshScr7() ********************************************/
/** Function to be called to update SCREEN 7.               **/
/*************************************************************/
void RefreshLine7(register byte Y)
{
  register byte X,C;
  register byte *T,*R;
  byte ZBuf[304];

  if (imgdepth != 16) {
    register byte *P;

    P=RefreshBorder(Y,XPal[BGColor]);
  
    if(!ScreenON) memset(P,XPal[BGColor],256);
    else
    {
      ColorSprites(Y,ZBuf);
      R=ZBuf+32;
      T=ChrTab+(((int)Y+VDP[23])<<8);
  
      for(X=0;X<32;X++)
      {
        C=R[0];P[0]=XPal[C? C:T[0]>>4];
        C=R[1];P[1]=XPal[C? C:T[1]>>4];
        C=R[2];P[2]=XPal[C? C:T[2]>>4];
        C=R[3];P[3]=XPal[C? C:T[3]>>4];
        C=R[4];P[4]=XPal[C? C:T[4]>>4];
        C=R[5];P[5]=XPal[C? C:T[5]>>4];
        C=R[6];P[6]=XPal[C? C:T[6]>>4];
        C=R[7];P[7]=XPal[C? C:T[7]>>4];
        R+=8;P+=8;T+=8;
      }
    }
  }
  else {
    register word *P;

    P=RefreshBorder16bpp(Y,XPal16bpp[BGColor]);
  
    if(!ScreenON) memset16bpp(P,XPal16bpp[BGColor],256);
    else
    {
      ColorSprites(Y,ZBuf);
      R=ZBuf+32;
      T=ChrTab+(((int)Y+VDP[23])<<8);
  
      for(X=0;X<32;X++)
      {
        C=R[0];P[0]=XPal16bpp[C? C:T[0]>>4];
        C=R[1];P[1]=XPal16bpp[C? C:T[1]>>4];
        C=R[2];P[2]=XPal16bpp[C? C:T[2]>>4];
        C=R[3];P[3]=XPal16bpp[C? C:T[3]>>4];
        C=R[4];P[4]=XPal16bpp[C? C:T[4]>>4];
        C=R[5];P[5]=XPal16bpp[C? C:T[5]>>4];
        C=R[6];P[6]=XPal16bpp[C? C:T[6]>>4];
        C=R[7];P[7]=XPal16bpp[C? C:T[7]>>4];
        R+=8;P+=8;T+=8;
      }
    }
  }
}

/** RefreshScr8() ********************************************/
/** Function to be called to update SCREEN 8.               **/
/*************************************************************/
void RefreshLine8(register byte Y)
{
  static byte SprToScr[16] =
  {
    0x00,0x02,0x10,0x12,0x80,0x82,0x90,0x92,
    0x49,0x4B,0x59,0x5B,0xC9,0xCB,0xD9,0xDB
  };
  register byte *T,*R;
  register byte C,X;
  byte ZBuf[304];

  if (imgdepth != 16) {
    register byte *P;

    P=RefreshBorder(Y,XPal[BGColor]);

    if(!ScreenON) memset(P,XPal[BGColor],256);
    else
    {
      ColorSprites(Y,ZBuf);
      R=ZBuf+32;
      Y+=VDP[23];
      T=ChrTab+((int)Y<<8);

      for(X=0;X<128;X++,T+=2,R+=2,P+=2)
      {
        C=R[0];P[0]=BPal[C? SprToScr[C]:T[0]];
        C=R[1];P[1]=BPal[C? SprToScr[C]:T[1]];
      }
    }
  }
  else {
    register word *P;

    P=RefreshBorder16bpp(Y,XPal16bpp[BGColor]);

    if (!ScreenON) memset16bpp(P,XPal16bpp[BGColor], 256);
    else {
      ColorSprites(Y,ZBuf);
      R=ZBuf+32;
      Y+=VDP[23];
      T=ChrTab+((int)Y<<8);

      for(X=0;X<128;X++,T+=2,R+=2,P+=2) {
        C=R[0];P[0]=BPal16bpp[C? SprToScr[C]:T[0]];
        C=R[1];P[1]=BPal16bpp[C? SprToScr[C]:T[1]];
      }
    }
  }
}

/** RefreshTx80() ********************************************/
/** Function to be called to update TEXT80.                 **/
/*************************************************************/
void RefreshLineTx80(register byte Y)
{
  register byte X,M;
  register byte *T,*C,*G;

  if (imgdepth != 16) {
    register byte FC,BC;
    register byte *P;

    P=RefreshBorder(Y,XPal[BGColor]);

    if(!ScreenON) memset(P,XPal[BGColor],256);
    else
    {
      G=(UseFont&&FontBuf? FontBuf:ChrGen)+(Y&0x07);
      T=ChrTab+80*(Y>>3);
      C=ColTab+10*(Y>>3);
      P-=32;
      M=0x80;

      for(X=0;X<80;X++,T++,P+=4)
      {
        Y=*(G+((int)*T<<3));
        if(*C&M) { FC=XPal[XFGColor];BC=XPal[XBGColor]; }
        else     { FC=XPal[FGColor];BC=XPal[BGColor]; }
        P[0]=Y&0xC0? FC:BC;
        P[1]=Y&0x30? FC:BC;
        P[2]=Y&0x0C? FC:BC;
        P[3]=BC;
        if(!(M>>=1)) { M=0x80;C++; } 
      }
    }
  }
  else {
    register word FC,BC;
    register word *P;

    P=RefreshBorder16bpp(Y,XPal16bpp[BGColor]);

    if(!ScreenON) memset16bpp(P,XPal16bpp[BGColor],256);
    else
    {
      G=(UseFont&&FontBuf? FontBuf:ChrGen)+(Y&0x07);
      T=ChrTab+80*(Y>>3);
      C=ColTab+10*(Y>>3);
      P-=32;
      M=0x80;

      for(X=0;X<80;X++,T++,P+=4)
      {
        Y=*(G+((int)*T<<3));
        if(*C&M) { FC=XPal16bpp[XFGColor];BC=XPal16bpp[XBGColor]; }
        else     { FC=XPal16bpp[FGColor];BC=XPal16bpp[BGColor]; }
        P[0]=Y&0xC0? FC:BC;
        P[1]=Y&0x30? FC:BC;
        P[2]=Y&0x0C? FC:BC;
        P[3]=BC;
        if(!(M>>=1)) { M=0x80;C++; } 
      }
    }
  }
}

#endif /* UNIX */
